#include <linux/limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif

#define KEYSZ	    8
#define MAXENTRIES 16
#define SECRETSZ   37

struct entry {
	char key[KEYSZ];
	char *value;
	int  sz;
};

int search(struct entry *e[], const char *key)
{
	int i;

	if (*key == ':')
		return -1;

	for (i = 0; i < MAXENTRIES; i++) {
		if (e[i] != NULL && strncmp(e[i]->key, key, KEYSZ) == 0)
			return i;
	}
	return -1;
}

int readsecret(char *dst, int sz);
int findempty(struct entry *e[]);
int readsz();
int readn(char *p, int n);
int writen(char *p, int n);

void child()
{
	struct entry *entries[MAXENTRIES];
	char searchkey[KEYSZ];
	char *ptr;
	char cmd;
	int i, sz;

	memset(entries, 0, sizeof(entries));

	entries[0] = malloc(sizeof(struct entry));
	strncpy(entries[0]->key, ":secret0", KEYSZ);
	entries[0]->sz = SECRETSZ;
	entries[0]->value = malloc(SECRETSZ);
	readsecret(entries[0]->value, SECRETSZ);

	while (read(0, &cmd, 1) > 0) {
		switch (cmd) {
		case '\n':
			break;
		case 'n': case 'p':
			i = findempty(entries);
			if (i < 0)
				break;
			entries[i] = malloc(sizeof(struct entry));
			memset(entries[i], 0, sizeof(struct entry));
			if (entries[i] == NULL)
				break;
			entries[i]->sz = readsz();
			if (entries[i]->sz <= 0)
				goto out;
			entries[i]->value = malloc(entries[i]->sz);
			if (entries[i]->value == NULL)
				goto out;
			ptr = &entries[i]->key[0];
			if (cmd == 'p')
				*ptr++ = ':';
			if (readn(ptr, KEYSZ) < 0)
				goto out;
			if (readn(entries[i]->value, entries[i]->sz) < 0)
				goto out;
			break;
		case 's':
			if (readn(searchkey, KEYSZ) < 0)
				goto out;
			i = search(entries, searchkey);
			if (i < 0)
				break;
			if (writen(entries[i]->value, entries[i]->sz) < 0)
				goto out;
			break;
		}
	}

out:
	for (i = 0; i < MAXENTRIES; i++) {
		if (entries[i] != NULL) {
			free(entries[i]->value);
			free(entries[i]);
			entries[i] = NULL;
		}
	}
}

/* NO INTENTIONAL ERRORS STARTING FROM THIS POINT */
int readsecret(char *dst, int sz)
{
	int fd, i;
	char c;

	fd = open("flag.txt", O_RDONLY);
	if (fd < 0)
		return -1;

	for (i = 0; i < sz; i++) {
		if (read(fd, dst, 1) < 0)
			return -1;
		dst++;
	}
	close(fd);
}


int findempty(struct entry *e[])
{
	int i;

	for (i = 0; i < MAXENTRIES; i++) {
		if (e[i] == NULL)
			return i;
	}
	return -1;
}

int readsz()
{
	int rv;
	char c;

	rv = 0;
	while (read(0, &c, 1) >= 0) {
		if (c < '0' || c > '9')
			break;
		rv += rv * 10 + (c - '0');
	}
	return rv;
}

int readn(char *p, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (read(0, p, 1) <= 0)
			return -1;
		p++;
	}
	return 0;
}

int writen(char *p, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (write(1, p, 1) != 1)
			return -1;
		p++;
	}
	return 0;
}


void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;
	char * volatile dummy;

	srand(time(NULL));

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			dummy = malloc(128);
			child();
			free(dummy);
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}
